@@ -114,8 +114,6 @@ void writecb(struct ev_loop *loop, ev_io *w, int revents) {
 int ClientHandler::noop() { return 0; }
 
 int ClientHandler::read_clear() {
-  ev_timer_again(conn_.loop, &conn_.rt);
-
   for (;;) {
     if (rb_.rleft() && on_read() != 0) {
       return -1;
@@ -124,6 +122,10 @@ int ClientHandler::read_clear() {
       rb_.reset();
     } else if (rb_.wleft() == 0) {
       conn_.rlimit.stopw();
+      if (reset_conn_rtimer_required_) {
+        reset_conn_rtimer_required_ = false;
+        ev_timer_again(conn_.loop, &conn_.rt);
+      }
       return 0;
     }
 
@@ -134,6 +136,10 @@ int ClientHandler::read_clear() {
     auto nread = conn_.read_clear(rb_.last, rb_.wleft());
 
     if (nread == 0) {
+      if (reset_conn_rtimer_required_) {
+        reset_conn_rtimer_required_ = false;
+        ev_timer_again(conn_.loop, &conn_.rt);
+      }
       return 0;
     }
 
@@ -208,8 +214,6 @@ int ClientHandler::tls_handshake() {
 }
 
 int ClientHandler::read_tls() {
-  ev_timer_again(conn_.loop, &conn_.rt);
-
   ERR_clear_error();
 
   for (;;) {
@@ -221,6 +225,11 @@ int ClientHandler::read_tls() {
       rb_.reset();
     } else if (rb_.wleft() == 0) {
       conn_.rlimit.stopw();
+      if (reset_conn_rtimer_required_) {
+        reset_conn_rtimer_required_ = false;
+        ev_timer_again(conn_.loop, &conn_.rt);
+      }
+
       return 0;
     }
 
@@ -231,6 +240,11 @@ int ClientHandler::read_tls() {
     auto nread = conn_.read_tls(rb_.last, rb_.wleft());
 
     if (nread == 0) {
+      if (reset_conn_rtimer_required_) {
+        reset_conn_rtimer_required_ = false;
+        ev_timer_again(conn_.loop, &conn_.rt);
+      }
+
       return 0;
     }
 
@@ -385,18 +399,14 @@ ClientHandler::ClientHandler(Worker *worker, int fd, SSL *ssl,
             get_config()->conn.upstream.ratelimit.write,
             get_config()->conn.upstream.ratelimit.read, writecb, readcb,
             timeoutcb, this, get_config()->tls.dyn_rec.warmup_threshold,
-            get_config()->tls.dyn_rec.idle_timeout),
-      pinned_http2sessions_(
-          get_config()->conn.downstream.proto == PROTO_HTTP2
-              ? make_unique<std::vector<ssize_t>>(
-                    worker->get_downstream_addr_groups().size(), -1)
-              : nullptr),
+            get_config()->tls.dyn_rec.idle_timeout, PROTO_NONE),
       ipaddr_(ipaddr),
       port_(port),
       faddr_(faddr),
       worker_(worker),
       left_connhd_len_(NGHTTP2_CLIENT_MAGIC_LEN),
-      should_close_after_write_(false) {
+      should_close_after_write_(false),
+      reset_conn_rtimer_required_(false) {
 
   ++worker_->get_worker_stat()->num_connections;
 
@@ -502,6 +512,10 @@ void ClientHandler::reset_upstream_write_timeout(ev_tstamp t) {
   }
 }
 
+void ClientHandler::signal_reset_upstream_conn_rtimer() {
+  reset_conn_rtimer_required_ = true;
+}
+
 int ClientHandler::validate_next_proto() {
   const unsigned char *next_proto = nullptr;
   unsigned int next_proto_len;
@@ -534,26 +548,26 @@ int ClientHandler::validate_next_proto() {
     return 0;
   }
 
+  auto proto = StringRef{next_proto, next_proto_len};
+
   if (LOG_ENABLED(INFO)) {
-    std::string proto(next_proto, next_proto + next_proto_len);
     CLOG(INFO, this) << "The negotiated next protocol: " << proto;
   }
 
-  if (!ssl::in_proto_list(get_config()->tls.npn_list, next_proto,
-                          next_proto_len)) {
+  if (!ssl::in_proto_list(get_config()->tls.npn_list, proto)) {
     if (LOG_ENABLED(INFO)) {
-      CLOG(INFO, this) << "The negotiated protocol is not supported";
+      CLOG(INFO, this) << "The negotiated protocol is not supported: " << proto;
     }
     return -1;
   }
 
-  if (util::check_h2_is_selected(next_proto, next_proto_len)) {
+  if (util::check_h2_is_selected(proto)) {
     on_read_ = &ClientHandler::upstream_http2_connhd_read;
 
     auto http2_upstream = make_unique<Http2Upstream>(this);
 
     upstream_ = std::move(http2_upstream);
-    alpn_.assign(next_proto, next_proto + next_proto_len);
+    alpn_.assign(std::begin(proto), std::end(proto));
 
     // At this point, input buffer is already filled with some bytes.
     // The read callback is not called until new data come. So consume
@@ -566,7 +580,7 @@ int ClientHandler::validate_next_proto() {
   }
 
 #ifdef HAVE_SPDYLAY
-  auto spdy_version = spdylay_npn_get_version(next_proto, next_proto_len);
+  auto spdy_version = spdylay_npn_get_version(proto.byte(), proto.size());
   if (spdy_version) {
     upstream_ = make_unique<SpdyUpstream>(spdy_version, this);
 
@@ -595,9 +609,9 @@ int ClientHandler::validate_next_proto() {
   }
 #endif // HAVE_SPDYLAY
 
-  if (next_proto_len == 8 && memcmp("http/1.1", next_proto, 8) == 0) {
+  if (proto == StringRef::from_lit("http/1.1")) {
     upstream_ = make_unique<HttpsUpstream>(this);
-    alpn_ = "http/1.1";
+    alpn_ = proto.str();
 
     // At this point, input buffer is already filled with some bytes.
     // The read callback is not called until new data come. So consume
@@ -642,13 +656,18 @@ void ClientHandler::pool_downstream_connection(
   if (!dconn->poolable()) {
     return;
   }
+
+  dconn->set_client_handler(nullptr);
+
+  auto group = dconn->get_downstream_addr_group();
+
   if (LOG_ENABLED(INFO)) {
     CLOG(INFO, this) << "Pooling downstream connection DCONN:" << dconn.get()
-                     << " in group " << dconn->get_group();
+                     << " in group " << group;
   }
-  dconn->set_client_handler(nullptr);
-  auto dconn_pool = worker_->get_dconn_pool();
-  dconn_pool->add_downstream_connection(std::move(dconn));
+
+  auto &dconn_pool = group->shared_addr->dconn_pool;
+  dconn_pool.add_downstream_connection(std::move(dconn));
 }
 
 void ClientHandler::remove_downstream_connection(DownstreamConnection *dconn) {
@@ -656,13 +675,14 @@ void ClientHandler::remove_downstream_connection(DownstreamConnection *dconn) {
     CLOG(INFO, this) << "Removing downstream connection DCONN:" << dconn
                      << " from pool";
   }
-  auto dconn_pool = worker_->get_dconn_pool();
-  dconn_pool->remove_downstream_connection(dconn);
+  auto &dconn_pool =
+      dconn->get_downstream_addr_group()->shared_addr->dconn_pool;
+  dconn_pool.remove_downstream_connection(dconn);
 }
 
 std::unique_ptr<DownstreamConnection>
 ClientHandler::get_downstream_connection(Downstream *downstream) {
-  size_t group;
+  size_t group_idx;
   auto &downstreamconf = get_config()->conn.downstream;
   auto catch_all = downstreamconf.addr_group_catch_all;
   auto &groups = worker_->get_downstream_addr_groups();
@@ -670,40 +690,43 @@ ClientHandler::get_downstream_connection(Downstream *downstream) {
   const auto &req = downstream->request();
 
   // Fast path.  If we have one group, it must be catch-all group.
-  // HTTP/2 and client proxy modes fall in this case.
+  // proxy mode falls in this case.
   if (groups.size() == 1) {
-    group = 0;
+    group_idx = 0;
   } else if (req.method == HTTP_CONNECT) {
     //  We don't know how to treat CONNECT request in host-path
     //  mapping.  It most likely appears in proxy scenario.  Since we
     //  have dealt with proxy case already, just use catch-all group.
-    group = catch_all;
+    group_idx = catch_all;
   } else {
     auto &router = get_config()->router;
+    auto &wildcard_patterns = get_config()->wildcard_patterns;
     if (!req.authority.empty()) {
-      group =
-          match_downstream_addr_group(router, StringRef{req.authority},
-                                      StringRef{req.path}, groups, catch_all);
+      group_idx =
+          match_downstream_addr_group(router, wildcard_patterns, req.authority,
+                                      req.path, groups, catch_all);
     } else {
       auto h = req.fs.header(http2::HD_HOST);
       if (h) {
-        group =
-            match_downstream_addr_group(router, StringRef{h->value},
-                                        StringRef{req.path}, groups, catch_all);
+        group_idx = match_downstream_addr_group(
+            router, wildcard_patterns, h->value, req.path, groups, catch_all);
       } else {
-        group =
-            match_downstream_addr_group(router, StringRef::from_lit(""),
-                                        StringRef{req.path}, groups, catch_all);
+        group_idx =
+            match_downstream_addr_group(router, wildcard_patterns, StringRef{},
+                                        req.path, groups, catch_all);
       }
     }
   }
 
   if (LOG_ENABLED(INFO)) {
-    CLOG(INFO, this) << "Downstream address group: " << group;
+    CLOG(INFO, this) << "Downstream address group_idx: " << group_idx;
   }
 
-  auto dconn_pool = worker_->get_dconn_pool();
-  auto dconn = dconn_pool->pop_downstream_connection(group);
+  auto &group = worker_->get_downstream_addr_groups()[group_idx];
+  auto &shared_addr = group.shared_addr;
+  auto &dconn_pool = shared_addr->dconn_pool;
+
+  auto dconn = dconn_pool.pop_downstream_connection();
 
   if (!dconn) {
     if (LOG_ENABLED(INFO)) {
@@ -711,22 +734,42 @@ ClientHandler::get_downstream_connection(Downstream *downstream) {
                        << " Create new one";
     }
 
-    auto dconn_pool = worker_->get_dconn_pool();
+    if (shared_addr->proto == PROTO_HTTP2) {
+      auto &http2_freelist = shared_addr->http2_freelist;
 
-    if (downstreamconf.proto == PROTO_HTTP2) {
-      Http2Session *http2session;
-      auto &pinned = (*pinned_http2sessions_)[group];
-      if (pinned == -1) {
-        http2session = worker_->next_http2_session(group);
-        pinned = http2session->get_index();
-      } else {
-        auto dgrp = worker_->get_dgrp(group);
-        http2session = dgrp->http2sessions[pinned].get();
+      if (http2_freelist.empty() ||
+          http2_freelist.size() < shared_addr->addrs.size()) {
+        if (LOG_ENABLED(INFO)) {
+          if (http2_freelist.empty()) {
+            CLOG(INFO, this)
+                << "http2_freelist is empty; create new Http2Session";
+          } else {
+            CLOG(INFO, this) << "Create new Http2Session; current "
+                             << http2_freelist.size() << ", min "
+                             << shared_addr->addrs.size();
+          }
+        }
+        auto session = make_unique<Http2Session>(
+            conn_.loop, shared_addr->tls ? worker_->get_cl_ssl_ctx() : nullptr,
+            worker_, &group);
+        http2_freelist.append(session.release());
       }
-      dconn = make_unique<Http2DownstreamConnection>(dconn_pool, http2session);
+
+      auto http2session = http2_freelist.head;
+
+      if (http2session->max_concurrency_reached(1)) {
+        if (LOG_ENABLED(INFO)) {
+          CLOG(INFO, this) << "Maximum streams are reached for Http2Session("
+                           << http2session
+                           << "). Remove Http2Session from http2_freelist";
+        }
+        http2_freelist.remove(http2session);
+      }
+
+      dconn = make_unique<Http2DownstreamConnection>(http2session);
     } else {
-      dconn = make_unique<HttpDownstreamConnection>(dconn_pool, group,
-                                                    conn_.loop, worker_);
+      dconn =
+          make_unique<HttpDownstreamConnection>(&group, conn_.loop, worker_);
     }
     dconn->set_client_handler(this);
     return dconn;
@@ -766,12 +809,6 @@ int ClientHandler::perform_http2_upgrade(HttpsUpstream *http) {
   auto downstream = http->get_downstream();
   auto input = downstream->get_response_buf();
 
-  static constexpr char res[] =
-      "HTTP/1.1 101 Switching Protocols\r\n"
-      "Connection: Upgrade\r\n"
-      "Upgrade: " NGHTTP2_CLEARTEXT_PROTO_VERSION_ID "\r\n"
-      "\r\n";
-
   if (upstream->upgrade_upstream(http) != 0) {
     return -1;
   }
@@ -784,7 +821,14 @@ int ClientHandler::perform_http2_upgrade(HttpsUpstream *http) {
   write_ = &ClientHandler::write_clear;
 
   input->remove(*output, input->rleft());
-  output->append(res, str_size(res));
+
+  constexpr auto res =
+      StringRef::from_lit("HTTP/1.1 101 Switching Protocols\r\n"
+                          "Connection: Upgrade\r\n"
+                          "Upgrade: " NGHTTP2_CLEARTEXT_PROTO_VERSION_ID "\r\n"
+                          "\r\n");
+
+  output->append(res);
   upstream_ = std::move(upstream);
 
   signal_write();
@@ -793,11 +837,11 @@ int ClientHandler::perform_http2_upgrade(HttpsUpstream *http) {
 
 bool ClientHandler::get_http2_upgrade_allowed() const { return !conn_.tls.ssl; }
 
-std::string ClientHandler::get_upstream_scheme() const {
+StringRef ClientHandler::get_upstream_scheme() const {
   if (conn_.tls.ssl) {
-    return "https";
+    return StringRef::from_lit("https");
   } else {
-    return "http";
+    return StringRef::from_lit("http");
   }
 }
 
@@ -811,24 +855,35 @@ namespace {
 // is mostly same routine found in
 // HttpDownstreamConnection::push_request_headers(), but vastly
 // simplified since we only care about absolute URI.
-std::string construct_absolute_request_uri(const Request &req) {
+StringRef construct_absolute_request_uri(BlockAllocator &balloc,
+                                         const Request &req) {
   if (req.authority.empty()) {
     return req.path;
   }
 
-  std::string uri;
+  auto len = req.authority.size() + req.path.size();
+  if (req.scheme.empty()) {
+    len += str_size("http://");
+  } else {
+    len += req.scheme.size() + str_size("://");
+  }
+
+  auto iov = make_byte_ref(balloc, len + 1);
+  auto p = iov.base;
+
   if (req.scheme.empty()) {
     // We may have to log the request which lacks scheme (e.g.,
     // http/1.1 with origin form).
-    uri += "http://";
+    p = util::copy_lit(p, "http://");
   } else {
-    uri += req.scheme;
-    uri += "://";
+    p = std::copy(std::begin(req.scheme), std::end(req.scheme), p);
+    p = util::copy_lit(p, "://");
   }
-  uri += req.authority;
-  uri += req.path;
+  p = std::copy(std::begin(req.authority), std::end(req.authority), p);
+  p = std::copy(std::begin(req.path), std::end(req.path), p);
+  *p = '\0';
 
-  return uri;
+  return StringRef{iov.base, p};
 }
 } // namespace
 
@@ -837,15 +892,17 @@ void ClientHandler::write_accesslog(Downstream *downstream) {
   const auto &req = downstream->request();
   const auto &resp = downstream->response();
 
+  auto &balloc = downstream->get_block_allocator();
+
   upstream_accesslog(
       get_config()->logging.access.format,
       LogSpec{
-          downstream, StringRef(ipaddr_), http2::to_method_string(req.method),
+          downstream, StringRef{ipaddr_}, http2::to_method_string(req.method),
 
           req.method == HTTP_CONNECT
               ? StringRef(req.authority)
-              : (get_config()->http2_proxy || get_config()->client_proxy)
-                    ? StringRef(construct_absolute_request_uri(req))
+              : get_config()->http2_proxy
+                    ? StringRef(construct_absolute_request_uri(balloc, req))
                     : req.path.empty()
                           ? req.method == HTTP_OPTIONS
                                 ? StringRef::from_lit("*")
@@ -974,23 +1031,23 @@ int ClientHandler::proxy_protocol_read() {
 
   --end;
 
-  constexpr char HEADER[] = "PROXY ";
+  constexpr auto HEADER = StringRef::from_lit("PROXY ");
 
-  if (static_cast<size_t>(end - rb_.pos) < str_size(HEADER)) {
+  if (static_cast<size_t>(end - rb_.pos) < HEADER.size()) {
     if (LOG_ENABLED(INFO)) {
       CLOG(INFO, this) << "PROXY-protocol-v1: PROXY version 1 ID not found";
     }
     return -1;
   }
 
-  if (!util::streq_l(HEADER, rb_.pos, str_size(HEADER))) {
+  if (!util::streq(HEADER, StringRef{rb_.pos, HEADER.size()})) {
     if (LOG_ENABLED(INFO)) {
       CLOG(INFO, this) << "PROXY-protocol-v1: Bad PROXY protocol version 1 ID";
     }
     return -1;
   }
 
-  rb_.drain(str_size(HEADER));
+  rb_.drain(HEADER.size());
 
   int family;
 
@@ -1125,18 +1182,18 @@ int ClientHandler::proxy_protocol_read() {
   return on_proxy_protocol_finish();
 }
 
-StringRef ClientHandler::get_forwarded_by() {
+StringRef ClientHandler::get_forwarded_by() const {
   auto &fwdconf = get_config()->http.forwarded;
 
   if (fwdconf.by_node_type == FORWARDED_NODE_OBFUSCATED) {
     return StringRef(fwdconf.by_obfuscated);
   }
 
-  return StringRef(faddr_->hostport);
+  return StringRef{faddr_->hostport};
 }
 
-const std::string &ClientHandler::get_forwarded_for() const {
-  return forwarded_for_;
+StringRef ClientHandler::get_forwarded_for() const {
+  return StringRef{forwarded_for_};
 }
 
 } // namespace shrpx
