@@ -48,6 +48,8 @@ namespace shrpx {
 
 class Http2DownstreamConnection;
 class Worker;
+struct DownstreamAddrGroup;
+struct DownstreamAddr;
 
 struct StreamData {
   StreamData *dlnext, *dlprev;
@@ -57,7 +59,7 @@ struct StreamData {
 class Http2Session {
 public:
   Http2Session(struct ev_loop *loop, SSL_CTX *ssl_ctx, Worker *worker,
-               size_t group, size_t idx);
+               DownstreamAddrGroup *group);
   ~Http2Session();
 
   // If hard is true, all pending requests are abandoned and
@@ -145,17 +147,31 @@ public:
 
   void submit_pending_requests();
 
-  const DownstreamAddr *get_addr() const;
+  DownstreamAddr *get_addr() const;
 
-  size_t get_group() const;
-
-  size_t get_index() const;
+  DownstreamAddrGroup *get_downstream_addr_group() const;
 
   int handle_downstream_push_promise(Downstream *downstream,
                                      int32_t promised_stream_id);
   int handle_downstream_push_promise_complete(Downstream *downstream,
                                               Downstream *promised_downstream);
 
+  // Returns number of downstream connections, including pushed
+  // streams.
+  size_t get_num_dconns() const;
+
+  // Returns true if this object is included in freelist.  See
+  // DownstreamAddrGroup object.
+  bool in_freelist() const;
+
+  // Returns true if the maximum concurrency is reached.  In other
+  // words, the number of currently participated streams in this
+  // session is equal or greater than the max concurrent streams limit
+  // advertised by server.  If |extra| is nonzero, it is added to the
+  // number of current concurrent streams when comparing against
+  // server initiated concurrency limit.
+  bool max_concurrency_reached(size_t extra = 0) const;
+
   enum {
     // Disconnected
     DISCONNECTED,
@@ -184,6 +200,8 @@ public:
 
   using ReadBuf = Buffer<8_k>;
 
+  Http2Session *dlnext, *dlprev;
+
 private:
   Connection conn_;
   DefaultMemchunks wb_;
@@ -203,13 +221,10 @@ private:
   Worker *worker_;
   // NULL if no TLS is configured
   SSL_CTX *ssl_ctx_;
+  DownstreamAddrGroup *group_;
   // Address of remote endpoint
-  const DownstreamAddr *addr_;
+  DownstreamAddr *addr_;
   nghttp2_session *session_;
-  size_t group_;
-  // index inside group, this is used to pin frontend to certain
-  // HTTP/2 backend for better throughput.
-  size_t index_;
   int state_;
   int connection_check_state_;
   bool flow_control_;
